@@ -25,8 +25,7 @@
 #include <linux/mm_inline.h>
 #include <linux/swap.h>
 #include <linux/writeback.h>
-#include <linux/buffer_head.h>
-#include <linux/module.h>
+#include <linux/export.h>
 #include <linux/syscalls.h>
 #include <linux/uio.h>
 #include <linux/security.h>
@@ -34,6 +33,7 @@
 #include <linux/socket.h>
 #include <net/sock.h>
 #include <linux/net.h>
+#include <linux/socket.h>
 #include <linux/genalloc.h>
 
 struct common_mempool;
@@ -701,12 +701,12 @@ static int pipe_to_sendpage(struct pipe_inode_info *pipe,
 
 	ret = buf->ops->confirm(pipe, buf);
 	if (!ret) {
-
 		more = (sd->flags & SPLICE_F_MORE) ? MSG_MORE : 0;
-		if (sd->len < sd->total_len)
+		if (sd->len < sd->total_len && pipe->nrbufs > 1)
 			more |= MSG_SENDPAGE_NOTLAST;
+
 		ret = file->f_op->sendpage(file, buf->page, buf->offset,
-				sd->len, &pos, more);
+					   sd->len, &pos, more);
 	}
 
 	return ret;
@@ -754,15 +754,12 @@ int pipe_to_file(struct pipe_inode_info *pipe, struct pipe_buffer *buf,
 		goto out;
 
 	if (buf->page != page) {
-		/*
-		 * Careful, ->map() uses KM_USER0!
-		 */
 		char *src = buf->ops->map(pipe, buf, 1);
-		char *dst = kmap_atomic(page, KM_USER1);
+		char *dst = kmap_atomic(page);
 
 		memcpy(dst + offset, src + buf->offset, this_len);
 		flush_dcache_page(page);
-		kunmap_atomic(dst, KM_USER1);
+		kunmap_atomic(dst);
 		buf->ops->unmap(pipe, buf, src);
 	}
 	ret = pagecache_write_end(file, mapping, sd->pos, this_len, this_len,
@@ -1443,7 +1440,7 @@ struct common_mempool
 };
 
 bool common_mempool_check_internal(common_mempool_t * pool,
-                                          void * ptr, 
+                                          void * ptr,
                                           common_mpool_hdr_t * hdr,
                                           common_mpool_ftr_t * ftr)
 {
@@ -1487,7 +1484,7 @@ void* common_mempool_alloc(common_mempool_t* pool)
 		return NULL;
 	}
 	spin_lock_bh(&pool->lock);
-	hdr = pool->head; 
+	hdr = pool->head;
 	pool->head = pool->head->next;
 
 	if (!pool->head) {
@@ -1800,7 +1797,7 @@ cleanup:
  */
 static int get_iovec_page_array(const struct iovec __user *iov,
 				unsigned int nr_vecs, struct page **pages,
-				struct partial_page *partial, int aligned,
+				struct partial_page *partial, bool aligned,
 				unsigned int pipe_buffers)
 {
 	int buffers = 0, error = 0;
@@ -2039,7 +2036,7 @@ static long vmsplice_to_pipe(struct file *file, const struct iovec __user *iov,
 		return -ENOMEM;
 
 	spd.nr_pages = get_iovec_page_array(iov, nr_segs, spd.pages,
-					    spd.partial, flags & SPLICE_F_GIFT,
+					    spd.partial, false,
 					    spd.nr_pages_max);
 	if (spd.nr_pages <= 0)
 		ret = spd.nr_pages;
@@ -2107,7 +2104,7 @@ SYSCALL_DEFINE6(splice, int, fd_in, loff_t __user *, off_in,
 	error = -EBADF;
 
 	/* check if fd_in is a socket */
-	sock = sockfd_lookup(fd_in, &error);
+	sock = sockfd_lookup(fd_in, (int *)&error);
 	if (sock) {
 		out = NULL;
 		if (!sock->sk)
@@ -2145,7 +2142,7 @@ done:
 	}
 
 	return error;
-	
+
 	}
 
 /*
@@ -2484,7 +2481,7 @@ static int __init init_splice_pools(void)
 {
 	unsigned int rcv_pool_size= sizeof(struct recvfile_ctl_blk) * MAX_PAGES_PER_RECVFILE;
 	unsigned int kve_pool_size= sizeof(struct kvec) * MAX_PAGES_PER_RECVFILE;
-	
+
 	rcv_pool =  common_mempool_create(4 * num_possible_cpus(), rcv_pool_size);
 	kvec_pool = common_mempool_create(4 * num_possible_cpus(), kve_pool_size);
 
@@ -2493,7 +2490,7 @@ static int __init init_splice_pools(void)
 		return -ENOMEM;
 	}
 /*
-	printk(KERN_ERR "%s rcv %p (sz:%d) kvec %p (sz:%d) per %d core\n", 
+	printk(KERN_ERR "%s rcv %p (sz:%d) kvec %p (sz:%d) per %d core\n",
 		__FUNCTION__, rcv_pool, rcv_pool_size, kvec_pool, kve_pool_size, num_possible_cpus());
 */
 }
